{$R-}  { This unit must be compiled with range checking turned off. }
UNIT Sorts;
{*******************************************************************}
{                                                                   }
{     WVS Software Company                                          }
{     Usage Fee: none, public domain                                }
{     Version: 1.0                                                  }
{     Release Date: 6/27/93                                         }
{                                                                   }
{     Programmer: Brad Williams                                     }
{     E-mail    : bwilliams@marvin.ag.uidaho.edu                    }
{     US Mail   : 1008 E. 7th                                       }
{                 Moscow, Idaho 83843                               }
{                                                                   }
{*******************************************************************}
{                                                                   }
{  This unit contains functions for performing various types of     }
{  sorts and searches on arrays.  To use any of the sort/search     }
{  routines, simply pass them an array and a compare or test        }
{  function.  Programs can be written to accept a TSortProcedure or }
{  TSearchFunction as a parameter and will then accept any sort or  }
{  search procedure provided in this unit.                          }
{                                                                   }
{****************************************************************************}
                                 INTERFACE
{****************************************************************************}
USES Objects;

TYPE
  TCompareFunction = FUNCTION (VAR AnArray; Item1, Item2 : LongInt) : Integer;
    { A TCompareFunction must return:   }
    {   1  if the Item1 > Item2         }
    {   0  if the Item1 = Item2         }
    {  -1  if the Item1 < Item2         }

  TSortProcedure = PROCEDURE  (VAR AnArray; Min, Max : LongInt;
                               Compare : TCompareFunction);

  TSwapProcedure  = PROCEDURE (VAR AnArray; Item1, Item2 : LongInt);


  { Sort Procedures that work with any array and any index range. }
PROCEDURE BinaryInsertionSort (VAR AnArray; Min, Max : LongInt;
                               Compare : TCompareFunction;
                               Swap : TSwapProcedure);
PROCEDURE BubbleSort (VAR AnArray;  Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);
PROCEDURE CombSort (VAR AnArray; Min, Max : LongInt;
                    Compare : TCompareFunction; Swap : TSwapProcedure);
PROCEDURE HeapSort (VAR APointerArray; Min, Max : LongInt;
                    Compare : TCompareFunction;
                    Swap : TSwapProcedure);
PROCEDURE QuickSort  (VAR APointerArray;  Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);
PROCEDURE QuickSortNonRecursive (VAR APointerArray; Min, Max : LongInt;
                                 Compare : TCompareFunction;
                                 Swap : TSwapProcedure);
PROCEDURE ShakerSort (VAR AnArray; Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);
PROCEDURE ShellSort  (VAR AnArray;  Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);
PROCEDURE StraightInsertionSort (VAR AnArray; Min, Max : LongInt;
                                 Compare : TCompareFunction;
                                 Swap : TSwapProcedure);
PROCEDURE StraightSelectionSort (VAR AnArray; Min, Max : LongInt;
                                 Compare : TCompareFunction;
                                 Swap : TSwapProcedure);

PROCEDURE TreeSort   (VAR APointerArray; Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);


  { Compare Procedures - Must write your own Compare for pointer variables. }
  { This allows one sort routine to be used on any array.                   }
FUNCTION  CompareChars    (VAR AnArray; Item1, Item2 : LongInt) : Integer;
                           FAR;
FUNCTION  CompareInts     (VAR AnArray; Item1, Item2 : LongInt) : Integer;
                           FAR;
FUNCTION  CompareLongInts (VAR AnArray; Item1, Item2 : LongInt) : Integer;
                           FAR;
FUNCTION  CompareReals    (VAR AnArray; Item1, Item2 : LongInt) : Integer;
                           FAR;
FUNCTION  CompareStrs     (VAR AnArray; Item1, Item2 : LongInt) : Integer;
                           FAR;

  { Swap procedures to be used in any sorting routine.  }
  { This allows one sorting routine to be on any array. }
PROCEDURE SwapChars    (VAR AnArray; A, B : LongInt); FAR;
PROCEDURE SwapInts     (VAR AnArray; A, B : LongInt); FAR;
PROCEDURE SwapLongInts (VAR AnArray; A, B : LongInt); FAR;
PROCEDURE SwapPtrs     (VAR AnArray; A, B : LongInt); FAR;
PROCEDURE SwapReals    (VAR AnArray; A, B : LongInt); FAR;
PROCEDURE SwapStrs     (VAR AnArray; A, B : LongInt); FAR;
{****************************************************************************}
                               IMPLEMENTATION
{****************************************************************************}
TYPE
  CharArray    = ARRAY[1..1] OF Char;
  IntArray     = ARRAY[1..1] OF Integer;
  LongIntArray = ARRAY[1..1] OF LongInt;
  PtrArray     = ARRAY[1..1] OF Pointer;
  RealArray    = ARRAY[1..1] OF Real;
  StrArray     = ARRAY[1..1] OF String;

{****************************************************************************}
{                                                                            }
{                      Local Procedures and Functions                        }
{                                                                            }
{****************************************************************************}
PROCEDURE AdjustArrayIndexes (VAR Min, Max : LongInt);
  { Adjusts array indexes to a one-based array. }
VAR Fudge : LongInt;
BEGIN
  Fudge := 1 - Min;
  Inc(Min,Fudge);
  Inc(Max,Fudge);
END;
{****************************************************************************}
{                                                                            }
{                      Global Procedures and Functions                       }
{                                                                            }
{****************************************************************************}
{                                                                            }
{                              Sort Procedures                               }
{                                                                            }
{****************************************************************************}
PROCEDURE BinaryInsertionSort (VAR AnArray; Min, Max : LongInt;
                               Compare : TCompareFunction;
                               Swap : TSwapProcedure);
  { Works with any array and any index range. }
VAR i, j, Middle, Left, Right : LongInt;
BEGIN
  AdjustArrayIndexes(Min,Max);
  FOR i := (Min + 1) TO Max DO
      BEGIN
        Left := Min;
        Right := i;
        WHILE Left < Right DO
          BEGIN
            Middle := (Left + Right) DIV 2;
            IF Compare(AnArray,Middle,i) < 1
               THEN Left := Middle + 1
               ELSE Right := Middle;
          END;
        FOR j := i DOWNTO (Right + 1) DO
            Swap(AnArray,j,j-1);
      END;
END;
{****************************************************************************}
PROCEDURE BubbleSort (VAR AnArray; Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);
  { Works with any array and any index range. }
VAR i, j : LongInt;
BEGIN
  AdjustArrayIndexes(Min,Max);
  FOR i := (Min + 1) TO Max DO
    FOR j := Max DOWNTO i DO
      IF Compare(AnArray,j-1,j) = 1
         THEN Swap(AnArray,j,j-1);
END;
{****************************************************************************}
PROCEDURE CombSort (VAR AnArray; Min, Max : LongInt;
                    Compare : TCompareFunction; Swap : TSwapProcedure);
  { The combsort is an optimised version of the bubble sort. It uses a }
  { decreasing gap in order to compare values of more than one element }
  { apart.  By decreasing the gap the array is gradually "combed" into }
  { order ... like combing your hair. First you get rid of the large   }
  { tangles, then the smaller ones ...                                 }
  {                                                                    }
  { There are a few particular things about the combsort. Firstly, the }
  { optimal shrink factor is 1.3 (worked out through a process of      }
  { exhaustion by the guys at BYTE magazine). Secondly, by never       }
  { having a gap of 9 or 10, but always using 11, the sort is faster.  }
  {                                                                    }
  { This sort approximates an n log n sort - it's faster than any      }
  { other sort I've seen except the quicksort (and it beats that too   }
  { sometimes ... have you ever seen a quicksort become an (n-1)^2     }
  { sort ... ?). The combsort does not slow down under *any*           }
  { circumstances. In fact, on partially sorted lists (including       }
  { *reverse* sorted lists) it speeds up.                              }
  {                                                                    }
  { More information in the April 1991 BYTE magazine.                  }
CONST ShrinkFactor = 1.3;
VAR Gap, i   : LongInt;
    Finished : Boolean;
BEGIN
  AdjustArrayIndexes(Min,Max);
  Gap := Round(Max/ShrinkFactor);
  REPEAT
    Finished := TRUE;
    Gap := Trunc(Gap/ShrinkFactor);
    IF Gap < 1
       THEN Gap := 1
       ELSE IF (Gap = 9) OR (Gap = 10)
               THEN Gap := 11;
    FOR i := Min TO (Max - Gap) DO
        IF Compare(AnArray,i,i+Gap) = 1
           THEN BEGIN
                  Swap(AnArray,i,i+Gap);
                  Finished := False;
                END;
  UNTIL ((Gap = 1) AND Finished);
END;
{****************************************************************************}
PROCEDURE HeapSort (VAR APointerArray; Min, Max : LongInt;
                    Compare : TCompareFunction; Swap : TSwapProcedure);
  { Performs best when items are in inverse order. }
VAR L, R : LongInt;
    X : Pointer;
    {*****************************************}
    PROCEDURE Sift;
    VAR i, j : LongInt;
        Label 13;
    BEGIN
      i := L;
      j := 2 * i;
      X := PtrArray(APointerArray)[i];
      WHILE j <= R DO
        BEGIN
          IF j < R
             THEN IF Compare(APointerArray,LongInt(PtrArray(APointerArray)[j]),j+1) = -1
                     THEN Inc(j);
          IF Compare(APointerArray,LongInt(X),j) >= 0
             THEN GoTo 13;
          PtrArray(APointerArray)[i] := PtrArray(APointerArray)[j];
          i := j;
          j := 2 * i;
        END;
      13: PtrArray(APointerArray)[i] := X;
    END;
    {*****************************************}
BEGIN
  AdjustArrayIndexes(Min,Max);
  L := (Max DIV 2) + 1;
  R := Max;
  WHILE L > 1 DO
    BEGIN
      Dec(L);
      Sift;
    END;
  WHILE R > 1 DO
    BEGIN
      X := PtrArray(APointerArray)[1];
      Swap(APointerArray,1,R);
      Dec(R);
      Sift;
    END;
END;
{****************************************************************************}
PROCEDURE QuickSort (VAR APointerArray; Min, Max : LongInt;
                     Compare : TCompareFunction; Swap : TSwapProcedure);
  {****************************************************************}
  PROCEDURE Sort (Left, Right : LongInt);
  VAR i, j  : LongInt;
      X : Pointer;
  BEGIN
    i := Left;
    j := Right;
    X := PtrArray(APointerArray)[(Left + Right) DIV 2];
    REPEAT
      WHILE Compare(APointerArray,LongInt(X),i) = 1 DO Inc(i);
      WHILE Compare(APointerArray,LongInt(X),j) = -1 DO Dec(j);
      IF i <= j
         THEN BEGIN
                Swap(APointerArray,i,j);
                Inc(i);
                Dec(j)
              END;
    UNTIL i > j;
    IF Left < j
       THEN Sort(Left,j);
    IF i < Right
       THEN Sort(i,Right)
  END;
  {****************************************************************}
BEGIN
  AdjustArrayIndexes(Min,Max);
  Sort(Min,Max);
END;
{****************************************************************************}
PROCEDURE QuickSortNonRecursive (VAR APointerArray; Min, Max : LongInt;
                                 Compare : TCompareFunction;
                                 Swap : TSwapProcedure);
CONST m = 12;
VAR i, j, L, R : LongInt;
    x : Pointer;
    s : 0..m;
    Stack : ARRAY[1..m] OF RECORD
                             l, r : LongInt;
                           END;
BEGIN
  AdjustArrayIndexes(Min,Max);
  s := 1;
  Stack[1].l := 1;
  Stack[1].r := max;
  REPEAT
    L := Stack[s].l;
    R := Stack[s].r;
    Dec(S);
    REPEAT
      i := L;
      j := R;
      x := PtrArray(APointerArray)[(L + R) DIV 2];
      REPEAT
        WHILE Compare(APointerArray,LongInt(x),i) = 1 DO Inc(i);
        WHILE Compare(APointerArray,LongInt(x),j) = -1 DO Dec(j);
        IF i <= j
           THEN BEGIN
                  Swap(APointerArray,i,j);
                  Inc(i);
                  Dec(j);
                END;
      UNTIL i > j;
      IF i < R
         THEN BEGIN
                Inc(s);
                Stack[s].l := i;
                Stack[s].r := R;
              END;
      R := j;
    UNTIL L >= R;
  UNTIL s = 0;
END;
{****************************************************************************}
PROCEDURE ShakerSort (VAR AnArray; Min, Max : LongInt;
                      Compare : TCompareFunction; Swap : TSwapProcedure);
  { Works for any array and any index range. }
VAR j, k, Left, Right : LongInt;
BEGIN
  AdjustArrayIndexes(Min,Max);
  Left := Min + 1;
  Right := Max;
  k := Max;
  REPEAT
    FOR j := Right DOWNTO Left DO
        IF Compare(AnArray,j-1,j) = 1
           THEN BEGIN
                  Swap(AnArray,j,j-1);
                  k := j;
                END;
    Left := k + 1;
    FOR j := Left TO Right DO
        IF Compare(AnArray,j-1,j) = 1
           THEN BEGIN
                  Swap(AnArray,j,j-1);
                  k := j;
                END;
    Right := k - 1;
  UNTIL Left > Right;
END;
{****************************************************************************}
PROCEDURE ShellSort (VAR AnArray; Min, Max : LongInt;
                     Compare : TCompareFunction; Swap : TSwapProcedure);
  { Works for any array and any index range. }
VAR Gap, i, j, k : LongInt;
BEGIN
  AdjustArrayIndexes(Min,Max);
  Gap := Max DIV 2;
  WHILE (Gap > 0) DO
    BEGIN
      FOR i := Gap TO Max DO
          BEGIN
            j := i - Gap;
            WHILE (j > 0) DO
              BEGIN
                k := j + Gap;
                IF Compare(AnArray,j,k) < 1
                   THEN j := 0
                   ELSE Swap(AnArray,j,k);
                Dec(j,Gap);
              END;
          END;
      Gap := Gap DIV 2;
    END;
END;
{****************************************************************************}
PROCEDURE StraightInsertionSort (VAR AnArray; Min, Max : LongInt;
                                 Compare : TCompareFunction;
                                 Swap : TSwapProcedure);
VAR i, j, k : LongInt;
    x : Pointer;
BEGIN
  AdjustArrayIndexes(Min,Max);
  FOR i := (Min + 1) TO Max DO
      BEGIN
        x := PtrArray(AnArray)[i];
        j := i;
        WHILE (j > 1) AND (Compare(AnArray,LongInt(x),j-1) = -1) DO
          BEGIN
            PtrArray(AnArray)[j] := PtrArray(AnArray)[j-1];
            Dec(j);
          END;
        PtrArray(AnArray)[j] := x;
      END;
END;
{****************************************************************************}
PROCEDURE StraightSelectionSort (VAR AnArray; Min, Max : LongInt;
                                 Compare : TCompareFunction;
                                 Swap : TSwapProcedure);
VAR i, j, k  : LongInt;
BEGIN
  AdjustArrayIndexes(Min,Max);
  FOR i := Min TO Max DO
      BEGIN
        k := i;
        FOR j := (i + 1) TO Max DO
            IF Compare(AnArray,j,k) = -1
               THEN k := j;
        Swap(AnArray,i,k);
      END;
END;
{****************************************************************************}
PROCEDURE TreeSort (VAR APointerArray; Min, Max : LongInt;
                    Compare : TCompareFunction; Swap : TSwapProcedure);
{after D.Cooke, A.H.Craven, G.M.Clarke: Statistical Computing
 in Pascal, Publisher: Edward Arnold, London 1985 ISBN 0-7131-3545-X }
 { Not functional }
TYPE PNode = ^Node;
     Node = RECORD
              Value : Pointer;
              Left  : PNode;
              Right : PNode;
            END;
VAR  Add, Top : PNode;
     i    : LongInt;
    {**************************************}
    PROCEDURE MakeTree (VAR Node : PNode);
    BEGIN
      IF Node = NIL
         THEN Node := Add
         ELSE IF Compare(APointerArray,LongInt(Add^.Value),
                                       LongInt(Node^.Value)) = 1
                 THEN MakeTree(Node^.Right)
                 ELSE MakeTree(Node^.Left);
    END;
    {**************************************}
     PROCEDURE StripTree (Node : PNode);
     BEGIN
       IF Node <> NIL
          THEN BEGIN
                 StripTree(Node^.Left);
                 PtrArray(APointerArray)[i] := Node^.Value;
                 Inc(i);
                 StripTree(Node^.Right)
               END;
     END;
    {**************************************}
BEGIN
  AdjustArrayIndexes(Min,Max);
  Top := NIL;
  FOR i := Min TO Max DO
    BEGIN
      New(Add);
      Add^.Value := PtrArray(APointerArray)[i];
      Add^.Left  := NIL;
      Add^.Right := NIL;
      MakeTree(Top)
    END;
    i := Min;
    StripTree(Top)
END;
{****************************************************************************}
{                                                                            }
{                            Compare Procedures                              }
{                                                                            }
{****************************************************************************}
FUNCTION CompareChars (VAR AnArray; Item1, Item2 : LongInt) : Integer;
BEGIN
  IF CharArray(AnArray)[Item1] < CharArray(AnArray)[Item2]
     THEN CompareChars := -1
     ELSE IF CharArray(AnArray)[Item1] = CharArray(AnArray)[Item2]
             THEN CompareChars := 0
             ELSE CompareChars := 1;
END;
{*****************************************************************************}
FUNCTION CompareInts (VAR AnArray; Item1, Item2 : LongInt) : Integer;
BEGIN
  IF IntArray(AnArray)[Item1] < IntArray(AnArray)[Item2]
     THEN CompareInts := -1
     ELSE IF IntArray(AnArray)[Item1] = IntArray(AnArray)[Item2]
             THEN CompareInts := 0
             ELSE CompareInts := 1;
END;
{*****************************************************************************}
FUNCTION CompareLongInts (VAR AnArray; Item1, Item2 : LongInt) : Integer;
BEGIN
  IF LongIntArray(AnArray)[Item1] < LongIntArray(AnArray)[Item2]
     THEN CompareLongInts := -1
     ELSE IF LongIntArray(AnArray)[Item1] = LongIntArray(AnArray)[Item2]
             THEN CompareLongInts := 0
             ELSE CompareLongInts := 1;
END;
{*****************************************************************************}
FUNCTION CompareReals (VAR AnArray; Item1, Item2 : LongInt) : Integer;
BEGIN
  IF RealArray(AnArray)[Item1] < RealArray(AnArray)[Item2]
     THEN CompareReals := -1
     ELSE IF RealArray(AnArray)[Item1] = RealArray(AnArray)[Item2]
             THEN CompareReals := 0
             ELSE CompareReals := 1;
END;
{*****************************************************************************}
FUNCTION CompareStrs (VAR AnArray; Item1, Item2 : LongInt) : Integer;
BEGIN
  IF StrArray(AnArray)[Item1] < StrArray(AnArray)[Item2]
     THEN CompareStrs := -1
     ELSE IF StrArray(AnArray)[Item1] = StrArray(AnArray)[Item2]
             THEN CompareStrs := 0
             ELSE CompareStrs := 1;
END;
{****************************************************************************}
{                                                                            }
{                             Move Procedures                                }
{                                                                            }
{****************************************************************************}
PROCEDURE MoveChar (VAR AnArray; Item : LongInt; VAR Hold);
BEGIN
  Char(Hold) := CharArray(AnArray)[Item];
END;
{****************************************************************************}
{                                                                            }
{                           MoveBack Procedures                              }
{                                                                            }
{****************************************************************************}
PROCEDURE MoveBackChar (VAR AnArray; Item : LongInt; VAR Hold);
BEGIN
  CharArray(AnArray)[Item] := Char(Hold);
END;
{****************************************************************************}
{                                                                            }
{                             Swap Procedures                                }
{                                                                            }
{****************************************************************************}
PROCEDURE SwapChars (VAR AnArray; A, B : LongInt);
VAR Item : Char;
BEGIN
  Item := CharArray(AnArray)[A];
  CharArray(AnArray)[A] := CharArray(AnArray)[B];
  CharArray(AnArray)[B] := Item;
END;
{*****************************************************************************}
PROCEDURE SwapInts (VAR AnArray; A, B : LongInt);
VAR Item : Integer;
BEGIN
  Item := IntArray(AnArray)[A];
  IntArray(AnArray)[A] := IntArray(AnArray)[B];
  IntArray(AnArray)[B] := Item;
END;
{*****************************************************************************}
PROCEDURE SwapLongInts (VAR AnArray; A, B : LongInt);
VAR Item : LongInt;
BEGIN
  Item := LongIntArray(AnArray)[A];
  LongIntArray(AnArray)[A] := LongIntArray(AnArray)[B];
  LongIntArray(AnArray)[B] := Item;
END;
{****************************************************************************}
PROCEDURE SwapPtrs (VAR AnArray; A, B : LongInt);
VAR Item : Pointer;
BEGIN
  Item := PtrArray(AnArray)[A];
  PtrArray(AnArray)[A] := PtrArray(AnArray)[B];
  PtrArray(AnArray)[B] := Item;
END;
{****************************************************************************}
PROCEDURE SwapReals (VAR AnArray; A, B : LongInt);
VAR Item : Real;
BEGIN
  Item := RealArray(AnArray)[A];
  RealArray(AnArray)[A] := RealArray(AnArray)[B];
  RealArray(AnArray)[B] := Item;
END;
{*****************************************************************************}
PROCEDURE SwapStrs (VAR AnArray; A, B : LongInt);
VAR Item : String;
BEGIN
  Item := StrArray(AnArray)[A];
  StrArray(AnArray)[A] := StrArray(AnArray)[B];
  StrArray(AnArray)[B] := Item;
END;
{*****************************************************************************}
BEGIN
END.

